#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Scons interface.

Interaction with scons. Find the binary, and run it with a set of given
options.

"""

from nuitka import Options, Tracing, Utils

import os, sys, subprocess

def getSconsDataPath():
    return Utils.dirname( __file__ )

def getSconsInlinePath():
    return Utils.joinpath( getSconsDataPath(), "inline_copy" )

def getSconsBinaryCall():
    """ Return a way to execute Scons.

        Using potentially inline copy if no system Scons is available
        or if we are on Windows.
    """
    if Utils.isFile( "/usr/bin/scons" ):
        return [ "/usr/bin/scons" ]
    else:
        return [
            getPython2ExePath(),
            Utils.joinpath( getSconsInlinePath(), "bin", "scons.py" )
        ]

def _getPython2ExePathWindows():
    # Shortcuts for the default installation directories, to avoid going to
    # registry at all.

    if os.path.exists( r"c:\Python27\python.exe" ):
        return r"c:\Python27\python.exe"
    elif os.path.exists( r"c:\Python26\python.exe" ):
        return r"c:\Python26\python.exe"

    # Windows only code, pylint: disable=E0602,F0401
    try:
        import _winreg as winreg
    except ImportError:
        import winreg

    for search in ( "2.7", "2.6" ):
        for arch_key in 0, winreg.KEY_WOW64_32KEY, winreg.KEY_WOW64_64KEY:
            try:
                key = winreg.OpenKey(
                    winreg.HKEY_LOCAL_MACHINE,
                    r"SOFTWARE\Python\PythonCore\%s\InstallPath" % search,
                    0,
                    winreg.KEY_READ | arch_key
                )

                return os.path.join(
                    winreg.QueryValue( key, '' ),
                    "python.exe"
                )
            except WindowsError:
                pass


def getPython2ExePath():
    """ Find a way to call Python2. Scons needs it."""

    if Utils.python_version < 300:
        return sys.executable
    elif os.name == "nt":
        python_exe = _getPython2ExePathWindows()

        if python_exe is not None:
            return python_exe
        else:
            sys.exit( """\
Error, need to find Python2 executable under C:\\Python26 or \
C:\\Python27 to execute scons which is not Python3 compatible.""" )
    elif os.path.exists( "/usr/bin/python2" ):
        return "python2"
    else:
        return "python"


def runScons( options, quiet ):
    # For the scons file to find the static C++ files and include path. The
    # scons file is unable to use __file__ for the task.
    os.environ[ "NUITKA_SCONS" ] = getSconsDataPath()

    if os.name == "nt":
        # On Windows this Scons variable must be set by us.
        os.environ[ "SCONS_LIB_DIR" ] = Utils.joinpath(
            getSconsInlinePath(),
            "lib",
            "scons-2.3.0"
        )

        # Also, for MinGW we can avoid the user having to add the path if he
        # used the default path or installed it on the same drive by appending
        # to the PATH variable before executing scons.
        os.environ[ "PATH" ] += r";\MinGW\bin;C:\MinGW\bin"

    scons_command = getSconsBinaryCall()

    if quiet:
        scons_command.append( "--quiet" )

    scons_command += [
        # The scons file
        "-f",
        Utils.joinpath( getSconsDataPath(), "SingleExe.scons" ),

        # Parallel compilation.
        "--jobs",
        str( Options.getJobLimit() ),

        # Do not warn about deprecations of Scons
        "--warn=no-deprecated",
    ]

    # Option values to provide to scons.
    for key, value in options.items():
        scons_command += [
            key + "=" + value
        ]

    if Options.isShowScons():
        Tracing.printLine( "Scons command:", " ".join( scons_command ) )

    return 0 == subprocess.call( scons_command )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Variable trace objects.

Variable traces indicate the flow of variables and merges their versions for
the SSA (Single State Assignment) form being used in Nuitka.

Variable version can start out:

* Unknown (maybe initialized, maybe not
* Uninit (definitely not initialized, first version, or after "del" statement)
* Init (definitely initialized, e.g. parameter variables)
* Merge (result of diverging code paths)

"""

from logging import debug

class VariableTraceBase:
    def __init__( self, variable, version ):
        self.variable     = variable
        self.version      = version

        # List of references.
        self.usages = []

        # List of releases of the node.
        self.releases = []

        # If not None, this indicates the last usage, where the value was not
        # yet escaped. If it is 0, it escaped immediately. Escaping is a one
        # time action.
        self.escaped_at = None

    def isNode( self ):
        return False

    def getVariable( self ):
        return self.variable

    def getVersion( self ):
        return self.version

    def addUsage( self, ref_node ):
        self.usages.append( ref_node )

    def addRelease( self, release_node ):
        self.releases.append( release_node )

    def onValueEscape( self ):
        self.escaped_at = len( self.usages )

    def isEscaped( self ):
        return self.escaped_at is not None

    def getPotentialUsages( self ):
        return self.usages

    def getDefiniteUsages( self ):
        return self.usages

    def getReleases( self ):
        return self.releases

    def isAssignTrace( self ):
        return False

    def isUninitTrace( self ):
        return False

    def isUnknownTrace( self ):
        return False

    def isMergeTrace( self ):
        return False


class VariableUninitTrace( VariableTraceBase ):
    def __init__( self, variable, version ):
        VariableTraceBase.__init__(
            self,
            variable = variable,
            version  = version
        )

    def __repr__( self ):
        return "<VariableUninitTrace %s %d>" % (
            self.variable,
            self.version
        )

    def isUninitTrace( self ):
        return True

    def dump( self ):
        debug( "Trace of %s %d:", self.variable, self.version )
        debug( "  Starts out uninitialized" )

        for count, usage in enumerate( self.usages ):
            if count == self.escaped_at:
                debug( "  Escaped value" )

            debug( "  Used at %s", usage )


class VariableUnknownTrace( VariableTraceBase ):
    def __init__( self, variable, version ):
        VariableTraceBase.__init__(
            self,
            variable = variable,
            version  = version
        )

    def __repr__( self ):
        return "<VariableUnknownTrace %s %d>" % (
            self.variable,
            self.version
        )

    def dump( self ):
        debug( "Trace of %s %d:", self.variable, self.version )
        debug( "  Starts unknown" )

        for count, usage in enumerate( self.usages ):
            if count == self.escaped_at:
                debug( "  Escaped value" )

            debug( "  Used at %s", usage )

    def isUnknownTrace( self ):
        return True


class VariableAssignTrace( VariableTraceBase ):
    def __init__( self, assign_node, variable, version ):
        VariableTraceBase.__init__(
            self,
            variable = variable,
            version  = version
        )

        self.assign_node = assign_node

    def __repr__( self ):
        return "<VariableAssignTrace %s %d at %s>" % (
            self.variable,
            self.version,
            self.assign_node.getSourceReference()
        )

    def dump( self ):
        debug( "Trace of %s %d:", self.variable, self.version )

        for count, usage in enumerate( self.usages ):
            if count == self.escaped_at:
                debug( "  Escaped value" )

            debug( "  Used at %s", usage )

    def isAssignTrace( self ):
        return True

    def getAssignNode( self ):
        return self.assign_node


class VariableMergeTrace( VariableTraceBase ):
    def __init__( self, variable, version, trace_yes, trace_no ):
        assert trace_no is not trace_yes, ( variable, version, trace_no )

        VariableTraceBase.__init__(
            self,
            variable = variable,
            version  = version
        )

        self.trace_yes = trace_yes
        self.trace_no = trace_no

        self.forwarded = True

    def isMergeTrace( self ):
        return True

    def addUsage( self, ref_node ):
        if not self.usages:
            # Merging is usage.
            self.trace_yes.addUsage( self )
            if self.trace_no is not None:
                self.trace_no.addUsage( self )

        VariableTraceBase.addUsage( self, ref_node )

    def getPotentialUsages( self ):
        assert False

    def dump( self ):
        debug( "Trace of %s %d:", self.variable, self.version )
        debug( "  Merge of %s <-> %s", self.trace_yes, self.trace_no )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Nodes related to raising and making exceptions.

"""

from .NodeBases import (
    ExpressionChildrenHavingBase,
    StatementChildrenHavingBase,
    ExpressionMixin,
    NodeBase
)

class StatementRaiseException( StatementChildrenHavingBase ):
    kind = "STATEMENT_RAISE_EXCEPTION"

    named_children = ( "exception_type", "exception_value", "exception_trace", "exception_cause" )

    def __init__( self, exception_type, exception_value, exception_trace, exception_cause, source_ref ):
        if exception_type is None:
            assert exception_value is None

        if exception_value is None:
            assert exception_trace is None

        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "exception_type"  : exception_type,
                "exception_value" : exception_value,
                "exception_trace" : exception_trace,
                "exception_cause" : exception_cause
            },
            source_ref = source_ref
        )

        self.reraise_local = False
        self.reraise_finally = False

    getExceptionType = StatementChildrenHavingBase.childGetter( "exception_type" )
    getExceptionValue = StatementChildrenHavingBase.childGetter( "exception_value" )
    getExceptionTrace = StatementChildrenHavingBase.childGetter( "exception_trace" )
    getExceptionCause = StatementChildrenHavingBase.childGetter( "exception_cause" )

    def isReraiseException( self ):
        return self.getExceptionType() is None

    def isReraiseExceptionLocal( self ):
        assert self.isReraiseException()

        return self.reraise_local

    def isReraiseExceptionFinally( self ):
        assert self.isReraiseException()

        return self.reraise_finally

    def markAsReraiseLocal( self ):
        self.reraise_local = True

    def markAsReraiseFinally( self ):
        self.reraise_finally = True

    def isStatementAborting( self ):
        return True

    def needsLineNumber( self ):
        return not self.isReraiseException()

    def isImplicit( self ):
        return False

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getExceptionType(), allow_none = True )
        exception_type = self.getExceptionType()

        if exception_type is not None and exception_type.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = exception_type,
                node       = self
            )

            return result, "new_raise", "Explicit raise already raises implicitely building exception type"

        constraint_collection.onExpression( self.getExceptionValue(), allow_none = True )
        exception_value = self.getExceptionValue()

        if exception_value is not None and exception_value.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    exception_type,
                    exception_value
                )
            )

            return result, "new_node", "Explicit raise already raises implicitely building exception value"

        constraint_collection.onExpression( self.getExceptionTrace(), allow_none = True )
        exception_trace = self.getExceptionTrace()

        if exception_trace is not None and exception_trace.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    exception_type,
                    exception_value,
                    exception_trace
                )
            )

            return result, "new_raise", "Explicit raise already raises implicitely building exception traceback"

        constraint_collection.onExpression( self.getExceptionCause(), allow_none = True )
        exception_cause = self.getExceptionCause()

        if exception_cause is not None and exception_cause.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    exception_type,
                    exception_cause,
                )
            )

            return result, "new_raise", "Explicit raise already raises implicitely building exception cause"

        return self, None, None


class StatementRaiseExceptionImplicit( StatementRaiseException ):
    kind = "STATEMENT_RAISE_EXCEPTION_IMPLICIT"

    def isStatementRaiseException( self ):
        return True

    def isImplicit( self ):
        return True


class ExpressionRaiseException( ExpressionChildrenHavingBase ):
    """ This node type is only produced via optimization.

    CPython only knows exception raising as a statement, but often the raising
    of exceptions can be predicted to occur as part of an expression, which it
    replaces then.
    """

    kind = "EXPRESSION_RAISE_EXCEPTION"

    named_children = ( "exception_type", "exception_value" )

    def __init__( self, exception_type, exception_value, source_ref ):
        ExpressionChildrenHavingBase.__init__(
            self,
            values     = {
                "exception_type"  : exception_type,
                "exception_value" : exception_value
            },
            source_ref = source_ref
        )

    def willRaiseException( self, exception_type ):
        # Virtual method, pylint: disable=R0201,W0613

        # One thing is clear, it will raise. TODO: Match exception_type more closely if it
        # is predictable.
        if exception_type is BaseException:
            return True
        else:
            return False


    getExceptionType = ExpressionChildrenHavingBase.childGetter( "exception_type" )
    getExceptionValue = ExpressionChildrenHavingBase.childGetter( "exception_value" )

    def mayProvideReference( self ):
        return False

    def computeExpression( self, constraint_collection ):
        return self, None, None


class ExpressionBuiltinMakeException( ExpressionChildrenHavingBase ):
    kind = "EXPRESSION_BUILTIN_MAKE_EXCEPTION"

    named_children = ( "args", )

    def __init__( self, exception_name, args, source_ref ):
        ExpressionChildrenHavingBase.__init__(
            self,
            values     = {
                "args" : tuple( args ),
            },
            source_ref = source_ref
        )

        self.exception_name = exception_name

    def getDetails( self ):
        return { "exception_name" : self.exception_name }

    def getExceptionName( self ):
        return self.exception_name

    getArgs = ExpressionChildrenHavingBase.childGetter( "args" )

    def computeExpression( self, constraint_collection ):
        return self, None, None


class ExpressionCaughtExceptionTypeRef( NodeBase, ExpressionMixin ):
    kind = "EXPRESSION_CAUGHT_EXCEPTION_TYPE_REF"

    def __init__( self, source_ref ):
        NodeBase.__init__( self, source_ref = source_ref )

    def computeExpression( self, constraint_collection ):
        # TODO: Might be predictable based on the exception handler this is in.
        return self, None, None

    def mayHaveSideEffects( self ):
        # Referencing the expression type has no side effect
        return False


class ExpressionCaughtExceptionValueRef( NodeBase, ExpressionMixin ):
    kind = "EXPRESSION_CAUGHT_EXCEPTION_VALUE_REF"

    def __init__( self, source_ref ):
        NodeBase.__init__( self, source_ref = source_ref )

    def computeExpression( self, constraint_collection ):
        # TODO: Might be predictable based on the exception handler this is in.
        return self, None, None

    def mayHaveSideEffects( self ):
        # Referencing the expression type has no side effect
        return False

    def makeCloneAt( self, source_ref ):
        return ExpressionCaughtExceptionValueRef(
            source_ref = source_ref
        )


class ExpressionCaughtExceptionTracebackRef( NodeBase, ExpressionMixin ):
    kind = "EXPRESSION_CAUGHT_EXCEPTION_TRACEBACK_REF"

    def __init__( self, source_ref ):
        NodeBase.__init__( self, source_ref = source_ref )

    def computeExpression( self, constraint_collection ):
        return self, None, None

    def mayHaveSideEffects( self ):
        # Referencing the expression type has no side effect
        return False



#!/usr/bin/env python
#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from __future__ import print_function

import os, sys, subprocess, tempfile, shutil

# Go its own directory, to have it easy with path knowledge.
os.chdir( os.path.dirname( os.path.abspath( __file__ ) ) )

search_mode = len( sys.argv ) > 1 and sys.argv[1] == "search"

start_at = sys.argv[2] if len( sys.argv ) > 2 else None

if start_at:
    active = False
else:
    active = True

if "PYTHON" not in os.environ:
    os.environ[ "PYTHON" ] = sys.executable

def check_output(*popenargs, **kwargs):
    from subprocess import Popen, PIPE, CalledProcessError

    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd, output=output)
    return output

version_output = check_output(
    [ os.environ[ "PYTHON" ], "--version" ],
    stderr = subprocess.STDOUT
)

python_version = version_output.split()[1]

os.environ[ "PYTHONPATH" ] = os.getcwd()

print( "Using concrete python", python_version )

for filename in sorted( os.listdir( "." ) ):
    if not os.path.isdir( filename ) or filename.endswith( ".build" ):
        continue

    path = os.path.relpath( filename )

    if not active and start_at in ( filename, path ):
        active = True

    if active:
        expected_errors = [
            "module_exits", "main_raises", "main_raises2",
            "package_contains_main"
        ]

        # Allowed after Python3, packages need no more "__init__.py"

        if python_version < b"3.3":
            expected_errors.append( "package_missing_init" )

        if filename not in expected_errors:
            extra_flags = [ "expect_success" ]
        else:
            extra_flags = [ "expect_failure" ]

        if filename in ( "package_missing_init", "dash_import", "reimport_main" ):
            extra_flags.append( "ignore_stderr" )

        extra_flags.append( "remove_output" )

        if filename == "plugin_import":
            os.environ[ "NUITKA_EXTRA_OPTIONS" ] = "--recurse-all --recurse-directory=%s/some_package" % filename
        else:
            os.environ[ "NUITKA_EXTRA_OPTIONS" ] = "--recurse-all"

        print( "Consider output of recursively compiled program:", path )
        sys.stdout.flush()

        for filename_main in os.listdir( filename ):
            if filename_main.endswith( "Main.py" ):
                break

            if filename_main.endswith( "Main" ):
                break
        else:
            sys.exit( "Error, no file ends with 'Main.py' or 'Main' in %s, incomplete test case" % filename )

        command = [
            sys.executable,
            os.path.join( "..", "..", "bin", "compare_with_cpython" ),
            os.path.join( filename, filename_main ),
            "silent",
        ]
        command += extra_flags

        result = subprocess.call(
            command
        )

        if result == 2:
            sys.stderr.write( "Interruped, with CTRL-C\n" )
            sys.exit( 2 )

        if result != 0 and search_mode:
            print( "Error exit!", result )
            sys.exit( result )
    else:
        print( "Skipping", filename )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Module with functions to display a node tree.

Useful to getting an idea of what the internal representation of Nuitka is about a source
code.
"""

from nuitka import SourceCodeReferences, Utils

from PyQt4 import QtCore, QtGui, uic

import sys

# The API requires a signature, sometimes we don't use it, pylint: disable=R0201
# Also using private stuff from classes, probably ok, pylint: disable=W0212

class NodeTreeModelItem:
    def __init__( self, node, parent = None ):
        self.parent_treeitem = parent
        self.node = node

        self.children = None

    def appendChild( self, _item ):
        assert False

    def _children( self ):
        if self.children is None:
            self.children = [
                NodeTreeModelItem( child, self )
                for child in
                self.node.getVisitableNodes()
            ]

        return self.children

    def child( self, row ):
        return self._children()[ row ]

    def childCount( self ):
        return len( self._children() )

    def columnCount( self ):
        return 2

    def data( self, column ):
        if column == 0:
            result = self.node.getDescription()
        elif column == 1:
            result = self.node.getDetail()
        else:
            assert False

        return QtCore.QVariant( result )

    def parent( self ):
        return self.parent_treeitem

    def row( self ):
        return self.parent_treeitem._children().index( self ) if self.parent else 0

class NodeTreeModel( QtCore.QAbstractItemModel ):
    def __init__( self, root, parent = None ):
        QtCore.QAbstractItemModel.__init__( self, parent )

        self.root_node = root
        self.root_item = NodeTreeModelItem( root )

    def columnCount( self, _parent ):
        return self.root_item.columnCount()

    def data( self, index, role ):
        if not index.isValid():
            return QtCore.QVariant()

        if role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()

        item = index.internalPointer()

        return QtCore.QVariant( item.data( index.column() ) )

    def flags( self, index ):
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def headerData( self, section, orientation, role ):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            if section == 0:
                return QtCore.QVariant( "Node Type" )
            elif section == 1:
                return QtCore.QVariant( "Node Detail" )

            return self.root_item.data( section )

        return QtCore.QVariant()

    def index( self, row, column, parent ):
        if row < 0 or column < 0 or row >= self.rowCount( parent ) or column >= self.columnCount( parent ):
            return QtCore.QModelIndex()

        if not parent.isValid():
            parent = self.root_item
        else:
            parent = parent.internalPointer()

        child = parent.child( row )

        if child:
            return self.createIndex( row, column, child )
        else:
            return QtCore.QModelIndex()

    def parent( self, index ):
        if not index.isValid():
            return QtCore.QModelIndex()

        child = index.internalPointer()
        parent = child.parent()

        if parent == self.root_item:
            return QtCore.QModelIndex()

        return self.createIndex( parent.row(), 0, parent )

    def rowCount( self, parent ):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent = self.root_item
        else:
            parent = parent.internalPointer()

        return parent.childCount()

    def getNodeFromPath( self, tree_path ):
        tree_path = list( tree_path )

        current = self.root_node

        while tree_path:
            current = current.getVisitableNodes()[ tree_path[0] ]

            del tree_path[0]

        return current

    def getItemFromSourceRef( self, source_ref ):
        def check( item ):
            if item.node.getSourceReference() == source_ref:
                return item

            for child in item._children():
                result = check( child )

                if result is not None:
                    return result

        return check( self.root_item )


class InspectNodeTreeDialog( QtGui.QDialog ):
    def __init__( self, *args ):
        QtGui.QDialog.__init__( self, *args )

        ui_dir = Utils.dirname( __file__ )
        ui_filename = Utils.joinpath( ui_dir, "dialogs", "InspectPythonTree.ui" )

        uic.loadUi( ui_filename, self )

        self.treeview_nodes.setSelectionMode( self.treeview_nodes.SingleSelection )

        self.displayed = None
        self.source_code = None
        self.model = None
        self.moving = None

    def setModel( self, model ):
        self.treeview_nodes.setModel( model )
        self.treeview_nodes.expandAll()

    @QtCore.pyqtSignature("on_treeview_nodes_clicked(QModelIndex)")
    def onTreeviewNodesClicked( self, item ):
        tree_path = []

        while item.isValid():
            tree_path.insert( 0, item.row() )

            item = item.parent()

        clicked_node = self.model.getNodeFromPath( tree_path )
        source_ref = clicked_node.getSourceReference()

        self.moving = True

        self.textedit_source.moveCursor( 1, 0 )

        for _i in range( 1, source_ref.getLineNumber()  ):
            self.textedit_source.moveCursor( 12, 0 )

        self.textedit_source.setFocus()

        self.moving = False

    @QtCore.pyqtSignature( "on_textedit_source_cursorPositionChanged()")
    def onTexteditSourceCursorMoved( self ):
        if self.moving:
            return

        pos = self.textedit_source.textCursor().position()

        code = self.source_code[:pos]

        line = 1

        for char in code:
            if char == "\n":
                line += 1

        # print "Line", line

        item = self.model.getItemFromSourceRef(
            self.displayed.atLineNumber(
                line = line
            )
        )

        if item is not None:
            item_path = []

            while item:
                item_path.insert( 0, item )

                item = item.parent()

            index = QtCore.QModelIndex()

            parent = self.model.root_item

            for item in item_path[1:]:
                index = index.child( parent._children().index( item )+1, 1 )
                parent = item

            # print self.treeview_nodes.visualRect( index )

    def loadSource( self, filename ):
        self.moving = True
        self.source_code = open( filename ).read()
        self.textedit_source.setPlainText( self.source_code  )
        self.moving = False

        self.displayed = SourceCodeReferences.fromFilename(
            filename    = filename,
            future_spec = None
        )


def displayTreeInspector( tree ):
    app = QtGui.QApplication( sys.argv )

    model = NodeTreeModel( tree )

    dialog = InspectNodeTreeDialog()
    dialog.setModel( model )
    dialog.model = model

    from . import SyntaxHighlighting

    SyntaxHighlighting.addPythonHighlighter(
        document = dialog.textedit_source.document()
    )
    dialog.loadSource( tree.getFilename() )

    dialog.setWindowFlags( QtCore.Qt.Window )
    dialog.show()

    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app.exec_()
